#include <cudnn.h>

#include <iostream>
#include <cstdlib>
#include <type_traits>
#include <vector>
#include <limits>
#include <random>
#include <algorithm>
#include <utility>
#include <thread>
#include <functional>
#include <ios>


#define checkCUDA(expression)                               \
  {                                                          \
    cudaError_t status = (expression);                     \
    if (status != cudaSuccess) {                    \
      std::cerr << "Error on line " << __LINE__ << ": "      \
                << cudaGetErrorString(status) << std::endl; \
      std::exit(EXIT_FAILURE);                               \
    }                                                        \
  }

#define checkCUDNN(expression)                               \
  {                                                          \
    cudnnStatus_t status = (expression);                     \
    if (status != CUDNN_STATUS_SUCCESS) {                    \
      std::cerr << "Error on line " << __LINE__ << ": "      \
                << cudnnGetErrorString(status) << std::endl; \
      std::exit(EXIT_FAILURE);                               \
    }                                                        \
  }


#define ASSERT(condition)                               \
{                                                          \
  if (!(condition)) {                                      \
    std::cerr << "ASSERT on line " << __LINE__ << '\n'; \
    std::exit(EXIT_FAILURE);                               \
  }                                                        \
}                                                                 

constexpr int BATCH_SIZE = 64;
constexpr int HIDDEN_SIZE = 384;
constexpr int INPUT_SIZE = 512;
constexpr int ITERATIONS = 1000;

constexpr int BENCH_ITERATIONS = 3000;

constexpr int N_STREAMS = 4;

using TensorType = float;

constexpr std::size_t PROJ_SIZE = HIDDEN_SIZE;

constexpr std::size_t X_SIZE = BATCH_SIZE * INPUT_SIZE;
constexpr std::size_t Y_SIZE = BATCH_SIZE * PROJ_SIZE;
constexpr std::size_t H_SIZE = BATCH_SIZE * HIDDEN_SIZE;
constexpr std::size_t W_SIZE = HIDDEN_SIZE * INPUT_SIZE;
constexpr std::size_t R_SIZE = HIDDEN_SIZE * HIDDEN_SIZE;
constexpr std::size_t B_SIZE = HIDDEN_SIZE;
constexpr std::size_t HO_SIZE = BATCH_SIZE * HIDDEN_SIZE;

constexpr std::size_t ELEMENT_SIZE = sizeof(TensorType);

constexpr std::size_t X_BYTES = X_SIZE * ELEMENT_SIZE;
constexpr std::size_t Y_BYTES = Y_SIZE * ELEMENT_SIZE;
constexpr std::size_t H_BYTES = H_SIZE * ELEMENT_SIZE;
constexpr std::size_t W_BYTES = W_SIZE * ELEMENT_SIZE;
constexpr std::size_t R_BYTES = R_SIZE * ELEMENT_SIZE;
constexpr std::size_t B_BYTES = B_SIZE * ELEMENT_SIZE;
constexpr std::size_t HO_BYTES = HO_SIZE * ELEMENT_SIZE;

constexpr std::size_t NUM_LAYERS = 1;
constexpr std::size_t LIN_LAYER_COUNT = 3;

static_assert(std::is_same<TensorType, float>::value);
constexpr cudnnDataType_t DATA_TYPE = CUDNN_DATA_FLOAT;

constexpr cudnnRNNDataLayout_t LAYOUT = CUDNN_RNN_DATA_LAYOUT_SEQ_MAJOR_PACKED;
constexpr int MAX_SEQ_LENGTH = 1;
constexpr void* PADDING_FILL = nullptr;

constexpr int NB_DIMS = 3;
constexpr cudnnForwardMode_t DNN_FORWARD_MODE = CUDNN_FWD_MODE_INFERENCE;

constexpr TensorType MIN = 0.0;
constexpr TensorType MAX = 10.0;


std::vector<int> SEQ_LENGTH_VEC(BATCH_SIZE);
constexpr std::size_t SEQ_LENGTH_BYTES = BATCH_SIZE * sizeof(int);

using Vec = std::vector<TensorType>;
using BiasVec = std::vector<Vec>;

template<typename Arr>
void fill_arr(
    Arr& arr,
    TensorType min = std::numeric_limits<TensorType>::min(),
    TensorType max = std::numeric_limits<TensorType>::max()) {
  static std::uniform_real_distribution<float> unif(min, max);
  static std::default_random_engine re;
  std::generate(arr.begin(), arr.end(), [&]() { return unif(re); });
}


struct HostConstants {
  HostConstants()
    : w{BiasVec(LIN_LAYER_COUNT)},
      r{BiasVec(LIN_LAYER_COUNT)},
      b{BiasVec(LIN_LAYER_COUNT + 1)} {
    
    for (int i = 0; i < LIN_LAYER_COUNT; ++i) {
      w[i] = Vec(W_SIZE);
      r[i] = Vec(R_SIZE);
      b[i] = Vec(B_SIZE);
    }
    b[3] = Vec(B_SIZE);

    fill();
  }

  void fill() {
    for (int i = 0; i < LIN_LAYER_COUNT; ++i) {
      fill_arr(w[i], MIN, MAX);
      fill_arr(r[i], MIN, MAX);
      fill_arr(b[i], MIN, MAX);
    }
    fill_arr(b[3], MIN, MAX);
  }

  BiasVec w;
  BiasVec r;
  BiasVec b;
};

struct HostTensors {
  HostTensors()
    : x{Vec(X_SIZE)},
      h{Vec(H_SIZE)} {
    // std::cout << "HostTensors()\n";
    fill();
  }

  void fill() {
    fill_arr(x);
    fill_arr(h);
  }

  Vec x;
  Vec h;
};

struct Inputs {
  Inputs(const HostTensors& tensors) {
    checkCUDA(cudaMalloc(&x_d, X_BYTES));
    checkCUDA(cudaMalloc(&h_d, H_BYTES));
    
    checkCUDA(cudaMemcpy(x_d, tensors.x.data(), X_BYTES, cudaMemcpyHostToDevice));
    checkCUDA(cudaMemcpy(h_d, tensors.h.data(), H_BYTES, cudaMemcpyHostToDevice));
  }

  ~Inputs() { 
    checkCUDA(cudaFree(h_d));
  }

  void* x_d;
  void* h_d;
};

struct Outputs {
  Outputs()
    : ho{Vec(HO_SIZE)} {
    // std::cout << "Outputs()\n";
    checkCUDA(cudaMalloc(&ho_d, HO_BYTES));
    checkCUDA(cudaMalloc(&y_d, Y_BYTES));
 }

  ~Outputs() {
    // std::cout << "~Outputs()\n";
    checkCUDA(cudaFree(y_d));
    checkCUDA(cudaFree(ho_d));
  }

  Vec ho;
  Vec y;

  void* ho_d;
  void* y_d;
};

struct DnnDescs {
  DnnDescs() {
    // std::cout << "DnnDescs()\n";
    checkCUDNN(cudnnCreateRNNDescriptor(&rnn_desc));
    const auto rnn_algo = CUDNN_RNN_ALGO_STANDARD;
    const auto rnn_mode = CUDNN_GRU;
    const auto bias_mode = CUDNN_RNN_DOUBLE_BIAS;
    const auto dir_mode = CUDNN_UNIDIRECTIONAL;
    const auto input_mode = CUDNN_LINEAR_INPUT;
    const auto math_prec = CUDNN_DATA_FLOAT;
    const auto math_type = CUDNN_TENSOR_OP_MATH;
    const cudnnDropoutDescriptor_t drop_out_desc = nullptr;
    const uint32_t aux_flags = CUDNN_RNN_PADDED_IO_DISABLED;
    checkCUDNN(cudnnSetRNNDescriptor_v8(
      rnn_desc,
      rnn_algo,
      rnn_mode,
      bias_mode,
      dir_mode,
      input_mode,
      DATA_TYPE,
      math_prec,
      math_type,
      INPUT_SIZE,
      HIDDEN_SIZE,
      PROJ_SIZE,
      NUM_LAYERS,
      drop_out_desc,
      aux_flags
    ));

    std::fill(SEQ_LENGTH_VEC.begin(), SEQ_LENGTH_VEC.end(), MAX_SEQ_LENGTH);

    checkCUDNN(cudnnCreateRNNDataDescriptor(&x_desc));
    const auto x_vector_size = INPUT_SIZE;
    checkCUDNN(cudnnSetRNNDataDescriptor(
      x_desc,
      DATA_TYPE,
      LAYOUT,
      MAX_SEQ_LENGTH,
      BATCH_SIZE,
      x_vector_size,
      SEQ_LENGTH_VEC.data(),
      PADDING_FILL
    ));

    checkCUDNN(cudnnCreateRNNDataDescriptor(&y_desc));
    const auto y_vector_size = PROJ_SIZE;
    checkCUDNN(cudnnSetRNNDataDescriptor(
      y_desc,
      DATA_TYPE,
      LAYOUT,
      MAX_SEQ_LENGTH,
      BATCH_SIZE,
      y_vector_size,
      SEQ_LENGTH_VEC.data(),
      PADDING_FILL
    ));

    checkCUDNN(cudnnCreateTensorDescriptor(&h_desc));
    const int h_dim_a[NB_DIMS] = {NUM_LAYERS, BATCH_SIZE, PROJ_SIZE};
    const int h_stride_a[NB_DIMS] = {BATCH_SIZE * PROJ_SIZE, PROJ_SIZE, 1};
    checkCUDNN(cudnnSetTensorNdDescriptor(h_desc, DATA_TYPE, NB_DIMS, h_dim_a, h_stride_a));

    cudnnHandle_t cudnn_handle;
    checkCUDNN(cudnnCreate(&cudnn_handle));
    
    weight_space_size = 0;
    checkCUDNN(cudnnGetRNNWeightSpaceSize(cudnn_handle, rnn_desc, &weight_space_size))
    ASSERT(weight_space_size >= LIN_LAYER_COUNT * (W_BYTES + R_BYTES + B_BYTES));

    work_space_size = 0;
    size_t reserve_space_size = 0;
    checkCUDNN(cudnnGetRNNTempSpaceSizes(
        cudnn_handle, rnn_desc, DNN_FORWARD_MODE, x_desc, &work_space_size, &reserve_space_size));
    ASSERT(reserve_space_size == 0);

    checkCUDNN(cudnnDestroy(cudnn_handle));
}

  ~DnnDescs() {
    // std::cout << "~DnnDescs()\n";
    checkCUDNN(cudnnDestroyTensorDescriptor(h_desc));
    checkCUDNN(cudnnDestroyRNNDataDescriptor(y_desc));
    checkCUDNN(cudnnDestroyRNNDataDescriptor(x_desc));
    checkCUDNN(cudnnDestroyRNNDescriptor(rnn_desc));
  }

  cudnnRNNDescriptor_t rnn_desc;
  cudnnRNNDataDescriptor_t x_desc;
  cudnnRNNDataDescriptor_t y_desc;
  cudnnTensorDescriptor_t h_desc;
  size_t weight_space_size;
  size_t work_space_size;
};

struct CommonBuffers {
  using DevBuffer = std::pair<void*, std::size_t>;

  CommonBuffers(const HostConstants& constants, const DnnDescs& descs) {
    // std::cout << "CommonBuffers()\n";
    checkCUDA(cudaMalloc(&seq_length_array, SEQ_LENGTH_BYTES));
    checkCUDA(cudaMalloc(&weight_space, descs.weight_space_size));  
    checkCUDA(cudaMemcpy(seq_length_array, SEQ_LENGTH_VEC.data(), SEQ_LENGTH_BYTES, cudaMemcpyHostToDevice));

    calc_weight_space(descs);
    init_weight_space(constants);
  }

  ~CommonBuffers() {
    // std::cout << "~CommonBuffers()\n";
    checkCUDA(cudaFree(weight_space));
    checkCUDA(cudaFree(seq_length_array));
  }

  void calc_weight_space(const DnnDescs& descs) {
    constexpr int32_t pseudo_layer = 0;

    w_dev_buffers.clear();
    r_dev_buffers.clear();
    b1_dev_buffers.clear();
    b2_dev_buffers.clear();
    size_t w_total_bytes = 0;
    size_t r_total_bytes = 0;
    size_t b1_total_bytes = 0;
    size_t b2_total_bytes = 0;

    cudnnHandle_t cudnn_handle;
    checkCUDNN(cudnnCreate(&cudnn_handle));
   
    int32_t lin_layer_id = 0;
    const int wb_nb_dims_requested = 3;
    cudnnDataType_t wb_data_type;
    int wb_nb_dims = 0;
    int wb_dim_a[wb_nb_dims_requested] = {};
    int wb_stride_a[wb_nb_dims_requested] = {};
    size_t tensor_size_bytes = 0;

    for (int i = 0; i < LIN_LAYER_COUNT; ++i) {
        lin_layer_id = i;
        cudnnTensorDescriptor_t w_desc;
        cudnnTensorDescriptor_t b1_desc;

        checkCUDNN(cudnnCreateTensorDescriptor(&w_desc));
        checkCUDNN(cudnnCreateTensorDescriptor(&b1_desc));

        void* w_addr = nullptr;
        void* b1_addr = nullptr;

        checkCUDNN(cudnnGetRNNWeightParams(cudnn_handle,
                                           descs.rnn_desc,
                                           pseudo_layer,
                                           descs.weight_space_size,
                                           weight_space,
                                           lin_layer_id,
                                           w_desc,
                                           &w_addr,
                                           b1_desc,
                                           &b1_addr));
        ASSERT(w_addr);
        checkCUDNN(cudnnGetTensorNdDescriptor(
            w_desc, wb_nb_dims_requested, &wb_data_type, &wb_nb_dims, wb_dim_a, wb_stride_a));
        ASSERT(wb_nb_dims == wb_nb_dims_requested);
        ASSERT(wb_data_type == DATA_TYPE);
        ASSERT(wb_dim_a[0] == 1 && wb_dim_a[1] == HIDDEN_SIZE && wb_dim_a[2] == INPUT_SIZE);
        ASSERT(wb_stride_a[0] == HIDDEN_SIZE * INPUT_SIZE && wb_stride_a[1] == INPUT_SIZE && wb_stride_a[2] == 1);
        checkCUDNN(cudnnGetTensorSizeInBytes(w_desc, &tensor_size_bytes));
        ASSERT(tensor_size_bytes >= HIDDEN_SIZE * INPUT_SIZE * ELEMENT_SIZE);
        w_dev_buffers.emplace_back(static_cast<uint8_t*>(w_addr), tensor_size_bytes);
        w_total_bytes += tensor_size_bytes;

        ASSERT(b1_addr);
        checkCUDNN(cudnnGetTensorNdDescriptor(
            b1_desc, wb_nb_dims_requested, &wb_data_type, &wb_nb_dims, wb_dim_a, wb_stride_a));
        ASSERT(wb_nb_dims == wb_nb_dims_requested);
        ASSERT(wb_data_type == DATA_TYPE);
        ASSERT(wb_dim_a[0] == 1 && wb_dim_a[1] == HIDDEN_SIZE && wb_dim_a[2] == 1);
        ASSERT(wb_stride_a[0] == HIDDEN_SIZE && wb_stride_a[1] == 1 && wb_stride_a[2] == 1);
        checkCUDNN(cudnnGetTensorSizeInBytes(b1_desc, &tensor_size_bytes));
        ASSERT(tensor_size_bytes >= HIDDEN_SIZE * ELEMENT_SIZE);
        b1_dev_buffers.emplace_back(static_cast<uint8_t*>(b1_addr), tensor_size_bytes);
        b1_total_bytes += tensor_size_bytes;

        checkCUDNN(cudnnDestroyTensorDescriptor(b1_desc));
        checkCUDNN(cudnnDestroyTensorDescriptor(w_desc));

        lin_layer_id = i + LIN_LAYER_COUNT;

        cudnnTensorDescriptor_t r_desc;
        cudnnTensorDescriptor_t b2_desc;

        checkCUDNN(cudnnCreateTensorDescriptor(&r_desc));
        checkCUDNN(cudnnCreateTensorDescriptor(&b2_desc));

        void* r_addr = nullptr;
        void* b2_addr = nullptr;

        checkCUDNN(cudnnGetRNNWeightParams(cudnn_handle,
                                           descs.rnn_desc,
                                           pseudo_layer,
                                           descs.weight_space_size,
                                           weight_space,
                                           lin_layer_id,
                                           r_desc,
                                           &r_addr,
                                           b2_desc,
                                           &b2_addr));
        ASSERT(r_addr);
        checkCUDNN(cudnnGetTensorNdDescriptor(
            r_desc, wb_nb_dims_requested, &wb_data_type, &wb_nb_dims, wb_dim_a, wb_stride_a));
        ASSERT(wb_nb_dims == wb_nb_dims_requested);
        ASSERT(wb_data_type == DATA_TYPE);
        ASSERT(wb_dim_a[0] == 1 && wb_dim_a[1] == HIDDEN_SIZE && wb_dim_a[2] == HIDDEN_SIZE);
        ASSERT(wb_stride_a[0] == HIDDEN_SIZE * HIDDEN_SIZE && wb_stride_a[1] == HIDDEN_SIZE && wb_stride_a[2] == 1);
        checkCUDNN(cudnnGetTensorSizeInBytes(r_desc, &tensor_size_bytes));
        ASSERT(tensor_size_bytes >= HIDDEN_SIZE * HIDDEN_SIZE * ELEMENT_SIZE);
        r_dev_buffers.emplace_back(static_cast<uint8_t*>(r_addr), tensor_size_bytes);
        r_total_bytes += tensor_size_bytes;

        if (b2_addr) {
          checkCUDNN(cudnnGetTensorNdDescriptor(
              b2_desc, wb_nb_dims_requested, &wb_data_type, &wb_nb_dims, wb_dim_a, wb_stride_a));
            checkCUDNN(cudnnGetTensorSizeInBytes(b2_desc, &tensor_size_bytes));
            b2_dev_buffers.emplace_back(static_cast<uint8_t*>(b2_addr), tensor_size_bytes);
            b2_total_bytes += tensor_size_bytes;
        }
        checkCUDNN(cudnnDestroyTensorDescriptor(b2_desc));
        checkCUDNN(cudnnDestroyTensorDescriptor(r_desc));
    }

    ASSERT(w_total_bytes >= LIN_LAYER_COUNT * W_BYTES &&
           r_total_bytes >= LIN_LAYER_COUNT * R_BYTES &&
           b1_total_bytes >= LIN_LAYER_COUNT * B_BYTES);
    ASSERT(descs.weight_space_size >= w_total_bytes + r_total_bytes + b1_total_bytes + b2_total_bytes);
  
    checkCUDNN(cudnnDestroy(cudnn_handle));
}

  void init_weight_space(const HostConstants& constants) {
    for (int i = 0; i < LIN_LAYER_COUNT; ++i) {
        ASSERT(w_dev_buffers[i].second >= W_BYTES);
        checkCUDA(cudaMemcpy(w_dev_buffers[i].first, constants.w[i].data(), W_BYTES, cudaMemcpyHostToDevice));

        ASSERT(r_dev_buffers[i].second >= R_BYTES);
        checkCUDA(cudaMemcpy(r_dev_buffers[i].first, constants.r[i].data(), R_BYTES, cudaMemcpyHostToDevice));

        ASSERT(b1_dev_buffers[i].second >= B_BYTES);
        checkCUDA(cudaMemcpy(b1_dev_buffers[i].first, constants.b[i].data(), B_BYTES, cudaMemcpyHostToDevice));

        if (i < b2_dev_buffers.size()) {
          if (i == 2) {
            checkCUDA(cudaMemcpy(b2_dev_buffers[i].first, constants.b[3].data(), B_BYTES, cudaMemcpyHostToDevice));
          } else {
            checkCUDA(cudaMemset(b2_dev_buffers[i].first, 0, b2_dev_buffers[i].second));
          }
        }
    }
  }

  int* seq_length_array;
  void* weight_space;

  std::vector<DevBuffer> w_dev_buffers;
  std::vector<DevBuffer> r_dev_buffers;
  std::vector<DevBuffer> b1_dev_buffers;
  std::vector<DevBuffer> b2_dev_buffers;
};

struct Buffers {
  explicit Buffers(std::size_t work_space_size) {
    // std::cout << "Buffers()\n";
    checkCUDA(cudaMalloc(&work_space, work_space_size));
  }

  ~Buffers() {
    // std::cout << "~Buffers()\n";
    checkCUDA(cudaFree(work_space));
  }

  void* work_space;
};

struct Context {
  Context() {
    // std::cout << "Context()\n";
    checkCUDNN(cudnnCreate(&handle));
    checkCUDA(cudaStreamCreate(&stream));
    checkCUDNN(cudnnSetStream(handle, stream));
  }

  ~Context() {
    // std::cout << "~Context()\n";
    checkCUDNN(cudnnDestroy(handle));
    checkCUDA(cudaStreamDestroy(stream));
  }

  cudnnHandle_t handle;
  cudaStream_t stream;
};


class Runner {
public:
  Runner() {
    HostConstants host_constants;
    std::vector<HostTensors> host_tensors(N_STREAMS);
    
    std::vector<Inputs> inputs;
    inputs.reserve(N_STREAMS);
    for (int i = 0; i < N_STREAMS; ++i) {
      inputs.emplace_back(host_tensors[i]);
    }
    std::vector<Outputs> outputs(N_STREAMS);

    DnnDescs descs;
    CommonBuffers common_buffers(host_constants, descs);
    
    std::vector<Buffers> buffers;
    buffers.reserve(N_STREAMS);
    for (int i = 0; i < N_STREAMS; ++i) {
      buffers.emplace_back(descs.work_space_size);
    }

    std::vector<Context> contexts(N_STREAMS);
    std::vector<std::thread> threads;
    threads.reserve(N_STREAMS);

    checkCUDA(cudaDeviceSynchronize());

    for (int i = 0; i < N_STREAMS; ++i) {
      threads.emplace_back(
        run_bench,
        std::cref(inputs[i]),
        std::ref(outputs[i]),
        std::cref(descs),
        std::cref(common_buffers),
        std::ref(buffers[i]),
        std::cref(contexts[i]));
    }

    for (auto& t : threads) {
      if (t.joinable()) {
          t.join();
      }
    }
}

  static void run_bench(
      const Inputs& inputs,
      Outputs& outputs,
      const DnnDescs& descs,
      const CommonBuffers& common_buffers,
      Buffers& buffers,
      const Context& context) {
    std::cout << "-----thread 0x" << std::hex << std::this_thread::get_id() << " run_bench()\n";
    for (int i = 0; i < BENCH_ITERATIONS; ++i) {
      run_sequence(inputs, outputs, descs, common_buffers, buffers, context);
      checkCUDA(cudaStreamSynchronize(context.stream));
    }
  }

  static void run_sequence(
      const Inputs& inputs,
      Outputs& outputs,
      const DnnDescs& descs,
      const CommonBuffers& common_buffers,
      Buffers& buffers,
      const Context& context) {
    for (int i = 0; i < ITERATIONS; ++i) {
      run(inputs, outputs, descs, common_buffers, buffers, context);
    }
  }

  static void run(
      const Inputs& inputs,
      Outputs& outputs,
      const DnnDescs& descs,
      const CommonBuffers& common_buffers,
      Buffers& buffers,
      const Context& context) {
    checkCUDNN(cudnnRNNForward(
        context.handle,
        descs.rnn_desc,
        DNN_FORWARD_MODE,
        common_buffers.seq_length_array,
        descs.x_desc,
        inputs.x_d,
        descs.y_desc,
        outputs.y_d,
        descs.h_desc,
        inputs.h_d,
        outputs.ho_d,
        nullptr,
        nullptr,
        nullptr,
        descs.weight_space_size,
        common_buffers.weight_space,
        descs.work_space_size,
        buffers.work_space,
        0,
        nullptr));
  }
};

int main(int argc, char const *argv[]) { 
  std::cout << "Test started\n";  
  Runner runner;
  std::cout << "Test passed\n";
  return EXIT_SUCCESS;
}